from typing import Generic, Optional, Tuple, TypeVar

from ..annotations import PyTree
from ..dataclass import dataclass

__all__ = ['GradientTransformation', 'GradientState']


GradientState = PyTree
Weights = TypeVar('Weights', bound=PyTree)
State = TypeVar('State', bound=PyTree)


@dataclass
class GradientTransformation(Generic[State, Weights]):

    T = TypeVar('T', bound='GradientTransformation[State, Weights]')

    def init(self, parameters: Weights) -> State:
        raise NotImplementedError

    def update(self,
               gradient: Weights,
               state: State,
               parameters: Optional[Weights]) -> Tuple[Weights, State]:
        raise NotImplementedError



from typing import Any

from .chain import ChainedGradientTransformation
from .transform import GradientTransformation
from .transforms import Scale, ScaleByAdam

__all__ = ['adam']


def adam(learning_rate: float,
         beta1: float = 0.9,
         beta2: float = 0.999,
         epsilon: float = 1e-8) -> GradientTransformation[Any, Any]:
    return ChainedGradientTransformation([ScaleByAdam(beta1, beta2, epsilon),
                                          Scale(-learning_rate)])



from typing import Any, Generic, List, Optional, Tuple, TypeVar

from ..annotations import PyTree
from ..dataclass import dataclass
from .transform import GradientTransformation, Weights

__all__ = ['ChainedGradientTransformation']


@dataclass
class ChainedGradientTransformation(GradientTransformation[List[PyTree], Weights],
                                    Generic[Weights]):

    U = TypeVar('U', bound='ChainedGradientTransformation[Weights]')

    transforms: List[GradientTransformation[Any, Weights]]

    def init(self, parameters: Weights) -> List[PyTree]:
        return [transform.init(parameters)
                for transform in self.transforms]

    def update(self,
               gradient: Weights,
               state: List[PyTree],
               parameters: Optional[Weights]) -> Tuple[Weights, List[PyTree]]:
        new_state: List[PyTree] = []
        for sub_state, transform in zip(state, self.transforms):
            gradient, new_sub_state = transform.update(gradient, sub_state, parameters)
            new_state.append(new_sub_state)
        return gradient, new_state
